﻿//------------------------------------------------------------------------------
//	合成音声を生成するために使うクラス
//		Resonator	共振器
//		Radiator	口からの放射の効果
//		Rosenberg	声帯音源として Rosenberg 波の発生
//------------------------------------------------------------------------------

using System;

namespace Mikami.FFT
{
	//	音声合成で使う共振器
	public class Resonator
	{
		private float piT_;
		private float a1_, a2_, b0_;
		private float yn1_, yn2_;

		public struct FrBw
		{
			public float fr_, bw_;
			public FrBw(float fr, float bw) { fr_ = fr; bw_ = bw; }
		}

		// 初期化を行うコンストラクタ
		public Resonator(float fr, float bw, float fs) { Initialize(fr, bw, fs); }

		// 初期化を行うコンストラクタ（構造体使用）
		public Resonator(FrBw fb, float fs) { Initialize(fb.fr_, fb.bw_, fs); }

		// 共振器のパラメータの設定
		public void Set(float fr, float bw)
		{
			a1_ = (float)(2.0*Math.Exp(-piT_*bw)*Math.Cos(2.0*piT_*fr));
			a2_ = (float)(-Math.Exp(-2.0*piT_*bw));
			b0_ = 1.0f - a1_ - a2_;
		}

		// 共振器のパラメータの設定（構造体使用）
		public void Set(FrBw fb) { Set(fb.fr_, fb.bw_); }

		// 内部の遅延器をクリア
		public void Clear() { yn1_ = 0; yn2_ = 0; }

		// 共振器に対応する処理の実行
		public float Execute(float xin)
		{
			float ym = a1_*yn1_ + a2_*yn2_ + b0_*xin;
			yn2_ = yn1_;	// 遅延器のデータの移動
			yn1_ = ym;		// 遅延器のデータの移動
			return ym;
		}

		// コンストラクタに共通な初期化
		private void Initialize(float fr, float bw, float fs)
		{
			piT_ = 3.14159265f/fs;
			Set(fr, bw);
			Clear();
		}
	}

	// 口からの放射の効果
	public class Radiator
	{
		private readonly float c1_;
		private float xnM1_;	

		public Radiator(float c = 1.0f)
		{
			c1_ = c;
			xnM1_ = 0;
		}

		public float Execute(float xin)
		{
			float yn = xin - c1_*xnM1_;
			xnM1_ = xin;	// x[n-1] ← x[n]
			return yn;
		}
	}

	// Rosenberg 波の発生
	public class Rosenberg
	{
		private readonly float dt_;
		private float period_, t1_, t2_, t_, amp_;
		
		public Rosenberg(float f0, float fs, float amp)
		{
			dt_ = 1.0f/fs;
			t_ = 0;
			amp_ = amp;
			SetPeriod(f0);
		}

		// Rosengerg 波の計算
		public float Execute()
		{
			float g = 0;

			if (t_ < t1_)
			{
				float x = t_/t1_;
				g = amp_*(3.0f - 2.0f*x)*x*x;
			}
			if ((t_ >= t1_) && (t_ < t1_+t2_))
			{
				float x = (t_ - t1_)/t2_;
				g = amp_*(1.0f - x*x);
			}

			if ((t_+=dt_) > period_) t_ -= period_;

			return g;
	}

		// 基本周期の設定
		public void SetPeriod(float f0)
		{
			period_ = 1.0f/f0;
			t1_ = 0.4f/f0;
			t2_ = 0.16f/f0;
		}

		// t_ = 0 に設定
		public void Reset() { t_ = 0; }	
	}
}
