﻿//------------------------------------------------------------------------------
//	FFT によるスペクトル解析 ― 窓掛け，零挿入を行う
//
//		作成者： 三上直樹， 2014/05/06 作成，(c)三上直樹
//------------------------------------------------------------------------------

using System;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;

using System.Numerics;
using System.Drawing.Drawing2D;
using Mikami.FFT;
using Mikami.WindowsApplication;

using Fb = Mikami.FFT.Resonator.FrBw;

namespace FFTスペクトル解析_窓掛け_ZeroPad
{
	public partial class Form1 : Form
	{
		private const float A0_ = -40;		// 振幅の最大値に対応するピクセル数
		private const float DT_ = 1.02f;	// 波形描画の 1 標本化間隔に対応するピクセル数
		private const float DF_ = 2.35f;	// 周波数の単位量に対応ピクセル数
		private const float DB1_ = -3.8f;	// 1 dB 当たりのピクセル数
		private const int N_FFT_ = 256;		// FFT の点数
		private const float FS_ = 8000;		// 標本化周波数

		Window windowing_;

		private Pen myPen0_ = new Pen(Color.Navy, 2);
		private Pen myPen1_ = new Pen(Color.CornflowerBlue, 2);
		private int nStart_;				// スペクトルを計算する際の最初の点
		private int nData_;					// 計算に使うデータ数
		private Resonator.FrBw[,] aiueo_ =
			{ { new Fb(654, 50), new Fb(1060, 55), new Fb(2375, 60) },	// ア
			  { new Fb(302, 40), new Fb(2057, 60), new Fb(3004, 65) },	// イ
			  { new Fb(375, 45), new Fb(1208, 55), new Fb(2165, 60) },	// ウ
			  { new Fb(541, 50), new Fb(1784, 60), new Fb(2452, 60) },	// エ
			  { new Fb(458, 45), new Fb( 807, 50), new Fb(2379, 60) }};	// オ
		private Resonator.FrBw[] c_ = new Resonator.FrBw[3];
		private const int N_RSN_ = 3;		// 共振器の数

		private PointF[] pt_, ptTrunc_, ptWindow_;
		private PointF[] ptSpc_;

		private FftReal myFft_ = new FftReal(N_FFT_);

		public Form1()
		{
			InitializeComponent();

			comboBox1.SelectedIndex = 0;	// 最初の画面では /ア/
			Paint -= Form1_Paint;
		}

		private void button1_Click(object sender, EventArgs e)
		{
			for (int n=0; n<3; n++)
				c_[n] = aiueo_[comboBox1.SelectedIndex, n];

			float[] sn;		// 標本化された合成音声
			
			VowelSynthesis(out sn);		// 母音合成
			if (sn == null) return;		// 入力した数値が不適切な場合はリターン

			// 波形描画用データの準備
			float snMax = sn.Max(x => Math.Abs(x));
			pt_ = sn.Select((x, n) => new PointF(n*DT_, (x/snMax)*A0_)).ToArray();

			//----------------------------------------------------------------------------
			// FFT によるスペクトル解析の計算
			double[] x1 = new double[N_FFT_];
			Complex[] y1 = new Complex[N_FFT_/2+1];

			// 差分フィルタ
			for (int n=0; n<nData_; n++) x1[n] = sn[n+nStart_] - sn[n-1+nStart_];
			// 零詰め
			for (int n=nData_; n<N_FFT_; n++) x1[n] = 0;

			// FFT で使う波形の表示の準備（窓掛け前）
			ptTrunc_ = x1.Select((x, n) => new PointF(n*DT_, ((float)x/snMax)*A0_)).ToArray();

			if (radioButton2.Checked) windowing_ = new Window(nData_, Window.Type.Hanning);
			if (radioButton3.Checked) windowing_ = new Window(nData_, Window.Type.Hamming);
			if (radioButton4.Checked) windowing_ = new Window(nData_, Window.Type.Blackman);

			// 窓掛け
			if (!radioButton1.Checked) windowing_.Execute(x1, x1);

			// FFT で使う波形の表示の準備（窓掛け後）
			ptWindow_ = x1.Select((x, n) => new PointF(n*DT_, ((float)x/snMax)*A0_)).ToArray();

			myFft_.Execute(x1, y1);		// FFT 実行

			var yAbs = y1.Select(x => (float)(Complex.Abs(x))).ToArray();
			float max = yAbs.Max();
			float[] db = yAbs.Select(x => (x > 0) ? (float)(20*Math.Log10(x/max)) : -100)
						.ToArray();
			ptSpc_ = db.Select((x, n) => new PointF(n*DF_, x*DB1_)).ToArray();

			float df = FS_/N_FFT_;
			textBox4.Clear();
			textBox4.AppendText(" [Hz] [dB]\r\n");
			for (int n=1; n<N_FFT_/2-1; n++)
			{
				if ((db[n] > -30.0f) && (db[n-1] < db[n]) && (db[n+1] < db[n]))
					textBox4.AppendText(String.Format(
						"{0,5:F0}  {1,3:F0}\r\n", n*df, db[n]));
			}

			//----------------------------------------------------------------------------

			Paint -= Form1_Paint;
			Paint += Form1_Paint;
			Invalidate();
		}

		private void Form1_Paint(object sender, PaintEventArgs e)
		{
			Graphics gr = e.Graphics;

			gr.TranslateTransform(20, 60);

			// 波形全体の描画
			DrawAxisX(gr, pt_.Length*DT_);
			gr.SmoothingMode = SmoothingMode.AntiAlias;
			gr.DrawCurve(myPen0_, pt_);

			// FFT の計算で使う波形の描画
			gr.ResetTransform();
			gr.TranslateTransform(20, 160);

			DrawAxisX(gr, N_FFT_*DT_);
			gr.SmoothingMode = SmoothingMode.AntiAlias;

			// 窓掛けの前の波形を描画
			gr.DrawCurve(myPen1_, ptTrunc_);	
			// 窓掛けの後の波形を描画
			gr.DrawCurve(myPen0_, ptWindow_);			 

			// スペクトルの描画
			gr.ResetTransform();
			gr.TranslateTransform(330, 160);
			gr.SmoothingMode = SmoothingMode.Default;

			float xMax = DF_*N_FFT_/2;
			float yMax = -DB1_*60;
			for (int n=0; n<=6; n++)
				gr.DrawLine(Pens.Black, 0, -10*n*DB1_, xMax, -10*n*DB1_);
			for (int n=0; n<=4; n++)
				gr.DrawLine(Pens.Black, n*xMax/4, 0, n*xMax/4, yMax);

			for (int n=0; n<=4; n++)
				gr.DrawString(n.ToString(), Font, Brushes.Black, n*xMax/4-4, yMax+6);
			gr.DrawString("周波数  [kHz]", Font, Brushes.Black, xMax/2-40, yMax+26);

			gr.DrawString("[dB]", Font, Brushes.Black, -22, -20);
			gr.DrawString("0", Font, Brushes.Black, -16, -5);
			for (int n=20; n<=60; n+=20)
				gr.DrawString((-n).ToString(), Font, Brushes.Black, -28, -DB1_*n-5);

			RectangleF rect = new RectangleF(0, 0, xMax, Math.Abs(yMax));

			gr.SetClip(rect);
			gr.SmoothingMode = SmoothingMode.AntiAlias;
			gr.DrawCurve(myPen0_, ptSpc_);
			gr.ResetClip();
		}

		// 波形の座標軸の描画
		private void DrawAxisX(Graphics gr, float length)
		{
			gr.SmoothingMode = SmoothingMode.Default;
			gr.DrawLine(Pens.Black, -5, 0, length+5, 0);

			const int Y0 = 40;
			for (int n=0; n<= (int)(length/DT_); n+=100)
				gr.DrawLine(Pens.Black, n*DT_, 4, n*DT_, -4);
			gr.DrawString("0", Font, Brushes.Black, -5, Y0);
			for (int n=100; n<= (int)(length/DT_); n+=100)
				gr.DrawString(n.ToString(), Font, Brushes.Black, n*DT_-13, Y0);
		}

		// 母音合成
		private void VowelSynthesis(out float[] sn)
		{

			sn = null;
			if (!textBox1.ToInt(out nStart_, x => (x >= 0) && (x <= 300),
								"0 ～ 300 の範囲にしてください．"))	return;
			float f0;
			if (!textBox2.ToFloat(out f0, f => (f >= 10) && (f <= 300),
								  "10 ～ 300 の範囲にしてください．")) return;
			if (!textBox3.ToInt(out nData_, x => (x >= 16) && (x <= N_FFT_),
								"16 ～ 256 の範囲にしてください．"))	return;

			Rosenberg g = new Rosenberg(f0, FS_, 1);	// 声帯波発生用オブジェクト初期化
													// 振幅：1
			Resonator[] rs = new Resonator[N_RSN_];	// 共振器用オブジェクト配列の宣言
			Radiator rd = new Radiator(1.0f);		// 放射の効果用オブジェクトの初期化

			for (int k=0; k<N_RSN_; k++)
				rs[k] = new Resonator(c_[k], FS_);

			const int N_DATA = 601;					// 標本化された合成音声のデータ数
			sn = new float [N_DATA];				// 標本化された合成音声

			// 合成音声生成
			for (int n=0; n<N_DATA; n++)
			{
				float vn = g.Execute();
				for (int k=0; k<N_RSN_; k++) vn = rs[k].Execute(vn);	// 声道の効果
				sn[n] = rd.Execute(vn);									// 放射の効果
			}
		}
	}
}
